//
//  CRF++ -- Yet Another CRF toolkit
//
//  $Id: encoder.cpp 1601 2007-03-31 09:47:18Z taku $;
//
//  Copyright(C) 2005-2007 Taku Kudo <taku@chasen.org>
//
#include <fstream>
#include "param.h"
#include "encoder.h"
#include "timer.h"
#include "tagger.h"
#include "lbfgs.h"
#include "common.h"
#include "feature_index.h"
#include "scoped_ptr.h"
#include "thread.h"
#include <assert.h>
#include <math.h>

#define WHAT_ERROR(msg) do {                                    \
  for (std::vector<TaggerImpl *>::iterator it = labelx.begin();    \
      it != labelx.end(); ++it)                                   \
  delete *it;                                               \
  std::cerr << msg << std::endl;                              \
  return false; } while (0)

namespace CRFPP {

class LabelLikelihoodThread: public thread {
    public:
      TaggerImpl **x;
      unsigned short start_i;
      unsigned short thread_num;
      int zeroone;
      int err;
      size_t size;
      double obj; //loglikelihood
      std::vector<double> expected; //gradient

      void run() {
        obj = 0.0;
        err = zeroone = 0;
        std::fill(expected.begin(), expected.end(), 0.0);
        for (size_t i = start_i; i < size; i += thread_num) {
          obj += x[i]->gradient(&expected[0]);
          int error_num = x[i]->eval();
          err += error_num;
          if (error_num) ++zeroone;
        }
      }
};


bool LabelLikelihood(const std::vector<TaggerImpl* > &x,
      EncoderFeatureIndex *feature_index, double *alpha,
      vector<double>& expected, double &value,
      size_t maxitr, double C, double eta,
      unsigned short thread_num, bool orthant)
{
    double old_obj = 1e+37;
    int    converge = 0;
    LBFGS lbfgs;
    std::vector<LabelLikelihoodThread> thread(thread_num);

    for (size_t i = 0; i < thread_num; i++) {
      thread[i].start_i = i;
      thread[i].size = x.size();
      thread[i].thread_num = thread_num;
      thread[i].x = const_cast<TaggerImpl **>(&x[0]);
      thread[i].expected.resize(feature_index->size());
    }

    size_t all = 0;
    for (size_t i = 0; i < x.size(); ++i)  all += x[i]->size();

    for (size_t itr = 0; itr < maxitr; ++itr) {
      for (size_t i = 0; i < thread_num; ++i) thread[i].start();
      for (size_t i = 0; i < thread_num; ++i) thread[i].join();

      for (size_t i = 1; i < thread_num; ++i) {
        thread[0].obj += thread[i].obj;
        thread[0].err += thread[i].err;
        thread[0].zeroone += thread[i].zeroone;
      }

      //combine gradient
      for (size_t i = 1; i < thread_num; ++i) {
        for (size_t k = 0; k < feature_index->size(); ++k)
          thread[0].expected[k] += thread[i].expected[k];
      }

      size_t num_nonzero = 0;
      if (orthant)
      {   // L1

      }
      else //currently focusing on L2
      {
		  for (size_t k = 0; k < feature_index->size(); ++k) {
			  if (k >= feature_index->sparse_feature_index) {
				thread[0].obj += (alpha[k] * alpha[k] /(2.0 * C));
				thread[0].expected[k] += alpha[k] / C;
			  }
		  }
      }

      double diff = (itr == 0 ? 1.0 :
          std::abs(old_obj - thread[0].obj)/old_obj);
      std::cout << "iter="  << itr
        << " obj=" << thread[0].obj
        << " diff="  << diff << std::endl;
      old_obj = thread[0].obj;

      if (diff < eta)
        converge++;
      else
        converge = 0;

      if (itr > maxitr || converge == 3)  break;  // 3 is ad-hoc

      if (lbfgs.optimize(feature_index->size(),
            &alpha[0],
            thread[0].obj,
            &thread[0].expected[0], orthant, C) <= 0)
        return false;
    }

    value = -thread[0].obj;

    return true;
  }

bool LBFGS_optimizer(const std::vector<TaggerImpl* > &x,
      EncoderFeatureIndex *feature_index, double *alpha,
      vector<double>& expected, double &value,
      size_t maxitr, double C, double eta, unsigned short shrinking_size,
      bool orthant)
{
    double old_obj = 1e+37;
    int    converge = 0;
    LBFGS lbfgs;

   expected.resize(feature_index->size());

   double obj = 0.0;
   // LBFGS loop
    for (size_t itr = 0; itr < maxitr; ++itr) {
    	    obj = 0.0;

    	    std::fill(expected.begin(), expected.end(), 0.0);
    	    // sum up gradient
		for (size_t i = 0; i < x.size(); i ++) {
			obj += x[i]->gradient(&expected[0]);
		}

		// L2 regularizer
		size_t num_nonzero = 0;
		for (size_t k = 0; k < feature_index->size(); ++k) {
			if (k >= feature_index->sparse_feature_index) {
				obj += (alpha[k] * alpha[k] /(2.0 * C));
				expected[k] += alpha[k] / C;
			}
		}

		double diff = (itr == 0 ? 1.0 :
        std::fabs(old_obj - obj)/old_obj);
		std::cout << "iter="  << itr
		<< " obj=" << obj
		<< " diff="  << diff << std::endl;
		old_obj = obj;

		if (diff < eta) converge++;
		else converge = 0;

		if (itr > maxitr || converge == 3)  break;  // 3 is ad-hoc

		if (lbfgs.optimize(feature_index->size(),
            &alpha[0],
            obj,
            &expected[0], orthant, C) <= 0)
        return false;
    }

    value = -obj;

    return true;
}

//----------------Learning using LBFGS-------------------
bool Encoder::learn(const char * modelfile)
{
	int ysize = feature_index->ysize();
	double tolerance = 0.001;
	double oldValue = 0.0;

	//---------for labeled instances------------
	//regular CRF optimization (compute loglikelihood and gradient, then LBFGS)
	vector<double> expected(feature_index->size());
	fill(expected.begin(), expected.end(), 0.0);
	double label_value = 0;

	// Multi-thread
	if (thread_num > 1) {
		LabelLikelihood(labelx, feature_index, &alpha[0], expected, label_value, maxitr, C, eta, thread_num, false);
	} else {
		// Single-thread
		LBFGS_optimizer(labelx, feature_index, &alpha[0], expected, label_value,
				maxitr, C, eta, shrinking_size, false);
	}

	if (!feature_index->save(modelfile, textmodel))
	WHAT_ERROR(feature_index->what());

	feature_index->freelist();

	std::cout << "\nDone!";

	return true;
}

bool Encoder::singleLabelLikelihood(TaggerImpl* x, EncoderFeatureIndex *feature_index,
	  double *alpha, vector<int> &alpha_count, vector<double> &expected,
	  double rate, double &value, double C, int n, int iter)
{
	  //store gradient for feature parameters
	  fill(expected.begin(), expected.end(), 0.0);

	  double obj = 0.0; //loglikelihoods
	  obj += x->gradient(&expected[0]);

#ifdef _DEBUG
	  cout<<"expected:"<<endl;
	  for (int i = 0; i < feature_index->size(); ++i) {
			cout<<expected[i]<<" ";
	  }
	  cout<<endl;
#endif

/*	  for (size_t k = 0; k < feature_index->size(); ++k) {
	  		  //obj += (alpha[k] * alpha[k] /(2.0 * C));
	  		  //update alpha
	  		  //alpha[k] += rate*(-expected[k]-alpha[k] / C);
	  		  double grad = -expected[k]-(double)alpha[k] / (n*C);
	  		  grad_alpha[k] += grad*grad;
	  		  if(grad_alpha[k] == 0)
	  		  	 continue;
	  		  alpha[k] += (double)rate/sqrt(grad_alpha[k])*grad;
	  		  //alpha[k] += rate*grad;
	  	  }
*/

	  for (set<int>::iterator it = x->feature_list.begin(); it != x->feature_list.end(); ++it) {
		  int k = *it;
		  double grad = -expected[k]-(double)alpha[k] / ((double)alpha_count[k]*C);
		  //double grad = -expected[k];
		  grad_alpha[k] += grad*grad;
		  if(grad_alpha[k] != 0) {
			  alpha[k] += (double)rate/sqrt(grad_alpha[k])*grad;
		  }
	  }

#ifdef _DEBUG
	  cout<<"alpha:"<<endl;
	  for (int i = 0; i < feature_index->size(); ++i) {
			cout<<alpha[i]<<" ";
	  }
	  cout<<endl;
#endif

	  value = -obj;
	  return true;
}

double Encoder::calculateAcc(const vector<TaggerImpl*> data)
{
	// read normal results
	double err = 0;
	int all = 0;
	for(int i=0; i<data.size(); i++) {
		data[i]->label_evaluate();

		err += data[i]->eval();
		all += data[i]->size();
	}

	return 1-(double)err/all;
}

//----------------Learning using SGD-------------------
bool Encoder::SGD_learn(string modelfile, double learningRate, double decayRate, int numIterations)
{
	// for debugging
	//highorder = false;
	int ysize = feature_index->ysize();
	bool converged = false;

	vector<double> expected(feature_index->size());
	fill(expected.begin(), expected.end(), 0);

	grad_alpha.resize(feature_index->size());
	fill(grad_alpha.begin(), grad_alpha.end(), 0);
	fill(alpha.begin(), alpha.end(), 0);

	alpha_count.resize(feature_index->size(), 0);
	for (int i = 0; i < labelx.size(); ++i) {
		for (set<int>::iterator it = labelx[i]->feature_list.begin(); it != labelx[i]->feature_list.end(); ++it) {
			alpha_count[*it] += 1;
		}
	}

    // Train index includes label + unlabel data.
	vector<int> trainindex;
	for (int i = 0; i < labelx.size(); i++)
		trainindex.push_back(i);

	int n = labelx.size();
	double lambda = 1.0 / n;
	double t = 1.0 / (lambda * learningRate);

    double eta = 1;
	double oldobj = 0;
	double unlabel_t = 0;

	cout<<"Start SGD iterations ... "<<endl;
	for(int iter=0; iter<numIterations; iter++) {
		Utils::shuffle(trainindex);

		double obj = 0.0;
		for (int i = 0; i < trainindex.size(); i++) {
			double label_value = 0;
			//eta = 1.0 / (lambda * t);
			//decaying learning rate
			if (!ada_grad) {
				eta = learningRate/(1.0+decayRate*unlabel_t);
			}

			//labelx[0]->PrintParameters();
			singleLabelLikelihood(labelx[trainindex[i]], feature_index, &alpha[0], alpha_count, expected, eta, label_value, C, trainindex.size(), iter);
			obj += label_value;
		}

		//labelx[0]->PrintMarginals();
		//labelx[0]->PrintParameters();

		double acc = calculateAcc(devx);
		cout<<"iter "<<iter<<" acc["<<iter<<"] = "<< acc<<" obj["<<iter<<"] = "<< obj<<endl;

		if (fabs(obj - oldobj) < 1e-3) {
			converged = true;
			break;
		}

		 oldobj = obj;
	}

	if(modelfile != "") {
		if (!feature_index->save(modelfile.c_str(), textmodel))
			WHAT_ERROR(feature_index->what());
	}
	std::cout << "\nDone!";

	feature_index->freelist();

	return true;
}

 void Encoder::RebuildAlpha() {
	feature_index->shrink(freq);

	alpha.resize(feature_index->size());           // parameter
	std::fill(alpha.begin(), alpha.end(), 0.0);
	feature_index->set_alpha(&alpha[0]);

	std::cout << "Number of features:  " << feature_index->size() << std::endl;
	std::cout << "Number of thread(s): " << thread_num << std::endl;
 }
}
