#include "feature_index.h"
#include "common.h"
#include "node.h"
#include "tagger.h"

#include "semi_tagger.h"
#include "../Parsing/Dictionary.h"
#include "../Parsing/Utils.h"

namespace CRFPP {
  static const char *BOS[4] = { "_B-1", "_B-2", "_B-3", "_B-4"};
  static const char *EOS[4] = { "_B+1", "_B+2", "_B+3", "_B+4"};

  const char* FeatureIndex::getFeature(int cur, int row, int col, TaggerImpl *tagger)
  {
	int idx = (int)cur + row;
	if (idx < 0)
	  return BOS[-idx-1];
	if (idx >= static_cast<int>(tagger->size()))
	  return EOS[idx - tagger->size()];

	return tagger->x(idx, col);
  }

   string getstrFeature(int cur, int row, int col, vector<vector<string> > &features)
   {
 	int idx = (int)cur + row;
 	if (idx < 0)
 	  return BOS[-idx-1];
 	if (idx >= static_cast<int>(features.size()))
 	  return EOS[idx - features.size()];

 	return features[idx][col];
   }

  void FeatureIndex::buildCRFTagger(TaggerImpl *tagger) {
      size_t fid = tagger->feature_id();
      size_t fvid = tagger->feature_value_id();

      unsigned short thread_id = tagger->thread_id();

      node_freelist_[thread_id].free();
      path_freelist_[thread_id].free();

      for (size_t cur = 0; cur < tagger->size(); ++cur) {
    	  	//cout<<"read feature cache "<<fid<<endl;
        int *f = feature_cache_[fid++];

        double *fv = NULL;
        if (fvid < feature_value_cache_.size())
        	fv = feature_value_cache_[fvid++];

        for (size_t i = 0; i < y_.size(); ++i) {
            Node *n = node_freelist_[thread_id].alloc();

			n->clear();
			n->x = cur;
			n->y = i;
			n->fvector = f;
			n->embedding_word_vec = fv;
			tagger->set_node(n, cur, i);
        }
      }

      if (!zero_order) {
    	  int *f = feature_cache_[fid++];

    	  for (size_t cur = 1; cur < tagger->size(); ++cur) {
			for (size_t j = 0; j < y_.size(); ++j) {
			  for (size_t i = 0; i < y_.size(); ++i) {
				Path *p = path_freelist_[thread_id].alloc();
				p->clear();
				p->add(tagger->node(cur-1, j), tagger->node(cur, i));
				p->fvector = f;
			  }
			}
		  }

//    	  int *edgef = feature_cache_[fid++];
//    	  int *emptyf = feature_cache_[fid++];
//		  for (size_t cur = 1; cur < tagger->size(); ++cur) {
//			for (size_t j = 0; j < y_.size(); ++j) {
//			  for (size_t i = 0; i < y_.size(); ++i) {
//				    Path *p = path_freelist_[thread_id].alloc();
//					//Path *p = new Path();
//					p->clear();
//					p->add(tagger->node(cur-1, j), tagger->node(cur, i));
//					if (transition_states.find(make_pair(j,i)) != transition_states.end()) {
//						p->fvector = edgef;
//				    } else {
//				    	p->fvector = emptyf;
//				    }
//			   }
//			}
//		  }
      }
    }

  void FeatureIndex::buildSemiCRFTagger(SemiTagger *tagger) {
	  unsigned short thread_id = tagger->thread_id();
	  size_t fid = tagger->feature_id();
	  size_t fvid = tagger->feature_value_id();

	  segnode_freelist_[thread_id].free();

	  //build features for all segment candidates
	  //all segments ending at position cur
	  //node[i][j][k]: the j-th segment ending at i with label k
	  for (size_t cur = 0; cur < tagger->size(); ++cur) {
		  int j = -1;
		  int start = -1;
		  while((start = tagger->nextSegStart_EndAt(cur, ++j)) != -1) {
			  int *f = feature_cache_[fid++];
			  double *fv = NULL;
			  if (fvid < feature_value_cache_.size())
				  fv = feature_value_cache_[fvid++];
			  int c = 0;
			  /*if (start == cur) {
				  SegmentNode *n = segnode_freelist_[thread_id].alloc();
				  n->y = nony;
				  n->fvector = f;
				  n->start = start;
				  n->end = cur;
				  tagger->set_node(n, cur, j, c++);
			  }*/
			  for (size_t i = 0; i < y_.size(); ++i) {
				  //if (i == nony) continue;
				  SegmentNode *n = segnode_freelist_[thread_id].alloc();
				  n->y = i;
				  n->fvector = f;
				  n->start = start;
				  n->end = cur;
				  n->embedding_word_vec = fv;
				  tagger->set_node(n, cur, j, c++);
			  }
		}
	  }

	//set node_beta
	//node_beta_[i][j][k]: the j-th node starting at i with label k
	for (size_t cur = 0; cur < tagger->size(); ++cur) {
		int j = -1;
		int end = -1;
		while((end = tagger->nextSegEnd_StartAt(cur, ++j)) != -1) {
			vector<SegmentNode*> nodes;
			tagger->getNodebySegmentEnd(cur, end, nodes); //segment [cur, end]
			for (size_t i = 0; i < nodes.size(); ++i) {
				tagger->set_node_beta(nodes[i], cur, j, i);
			}
		  }
	  }
  }

  void FeatureIndex::LoadDictFeatures(string filename) {
	  ifstream infile(filename.c_str(), ios::in);
	  string str;
	  while (getline(infile, str)) {
		  vector<string> values;
		  Utils::Split(str, '\t', values);
		  dict_features[values[0]] = 1;
	  }
	  infile.close();
  }

  void FeatureIndex::SetLabels(set<string> candset) {
   	y_.clear();

   	// make sure O corresponds to index 0
   	if (candset.find("O") != candset.end()) {
   		lb2i["O"] = y_.size();
   		i2lb[y_.size()] = "O";
   		nony = y_.size();
   		y_.push_back("O");
   	}

   	for (std::set<std::string>::iterator it = candset.begin(); it != candset.end(); ++it) {
   		if (*it == "O") continue;
   		lb2i[*it] = y_.size();
   		i2lb[y_.size()] = *it;
   		y_.push_back(this->strdup(it->c_str()));
   	}
   }

  bool FeatureIndex::buildSemiCRFFeatures(SemiTagger *tagger, bool train) {
	  	tagger->set_feature_id(feature_cache_.size());
	  	tagger->set_feature_value_id(feature_value_cache_.size());

	  	int id;
	  	vector<int> features;
	  	vector<double> feature_values;

	  	// only build dict features for the truesegments
	    for (size_t cur = 0; cur < tagger->size(); ++cur) {
	  		int j = -1;
	  		int start = -1;
	  		while((start = tagger->nextSegStart_EndAt(cur, ++j)) != -1) {
	  			vector<string> segfeatures;
	  			tagger->sent->genSegmentFeatures(start, cur, segfeatures);
	  			tagger->sent->genSegmentBoundaryFeatures(start, cur, segfeatures);

	  			tagger->sent->genSegmentSyntacticFeatures(start, cur, segfeatures);

	  			// segment features
	  			if (dict_features.size() > 0) {
					string segment = "";
					for (int s = start; s <= cur; ++s) {
						segment += tagger->sent->words[s] + "_";
					}
					if (dict_features.find(segment) != dict_features.end()) {
						segfeatures.push_back(segment);
					}
	  			}

	  			if (feature_map.size() > 0) {
	  				for (int s = start; s <= cur; ++s) {
	  				// feature map: subj, wordnet, verbnet
	  					string word = tagger->sent->words[s];
	  					string postag = tagger->sent->postags[s];
	  					string key = Utils::toLower(word) + "/" + postag[0];
						map<string, vector<string> >::iterator it = feature_map.find(key);
						if (it != feature_map.end()) {
						  for (int k = 0; k < it->second.size(); ++k) {
							  segfeatures.push_back(it->second[k]);
						  }
						}
					}
	  			}

	  			for(int i=0; i<segfeatures.size(); i++) {
	  				string fv = "F_"+ segfeatures[i];
	  				if (train) {
	  					id = getID(fv.c_str());
					} else {
						id = getfID(fv.c_str());
					}
	  				if(id != -1) {
	  					features.push_back(id);
	  				}
	  			}

	  			feature_cache_.add(features);
	  			features.clear();

	  			if (word_lookup_table.size() > 0 && word_lookup_table.size() > 0) {
					tagger->sent->genEmbeddingFeatures(start, cur,
								word_lookup_table, word_embeddings, feature_values);
					feature_value_cache_.add(feature_values);
					feature_values.clear();
	  			}
			 }
		  }

		if (!zero_order) {
			// edge features
			string fv = "B_";
			if (train) {
				edgeid_ = getID(fv.c_str());
			} else {
				edgeid_ = getfID(fv.c_str());
			}
		}
		return true;
  }

  bool FeatureIndex::buildCRFFeatures(TaggerImpl *tagger, bool train) {
         int id;
         string_buffer os;
         std::vector <int> feature;
         vector<double> feature_values;

         tagger->set_feature_id(feature_cache_.size());
         tagger->set_feature_value_id(feature_value_cache_.size());
         for (size_t cur = 0; cur < tagger->size(); ++cur) {
			vector<string> strfeatures;
	         // basic features: word, pos
			string word = tagger->sent->words[cur];
			string postag = tagger->sent->postags[cur];
			string key = Utils::toLower(word) + "/" + postag[0];
			tagger->sent->genBasicFeatures(cur, strfeatures);

			// window features, only consider word and lemma by default
			for (int k = 1; k <= windowsize; ++k) {
				tagger->sent->genCRFWindowFeatures(k, cur, strfeatures);
				tagger->sent->genCRFWindowFeatures(-1*k, cur, strfeatures);
			}

			 // more lexical features
			 if (feature_map.size() > 0) {
				 map<string, vector<string> >::iterator it = feature_map.find(key);
				if (it != feature_map.end()) {
				  for (int k = 0; k < it->second.size(); ++k) {
					  strfeatures.push_back(it->second[k]);
				  }
				}
			 }

			for(int col=0; col<strfeatures.size(); col++) {
				string fv = "F_"+strfeatures[col]; //w_-1_0_be
				if (train) {
					id = getID(fv.c_str());
				} else {
					id = getfID(fv.c_str());
				}
				if (id != -1)
					feature.push_back(id);
			}

     		feature_cache_.add(feature);
     		feature.clear();

     		// embedding features
     		if (word_lookup_table.size() > 0 && word_lookup_table.size() > 0) {
				tagger->sent->genEmbeddingFeatures(cur, cur, word_lookup_table, word_embeddings, feature_values);
				feature_value_cache_.add(feature_values);
				feature_values.clear();
     		}
         }

         if (!zero_order) {
  		    //default edge feature
  			string fv = "B_";
  			if (train) {
  				id = getID(fv.c_str());
  			} else {
  				id = getfID(fv.c_str());
  			}
  			if (id != -1)
  				feature.push_back(id);
  			feature_cache_.add(feature);
  			feature.clear();
        }
      return true;
    }

}


