//
//  CRF++ -- Yet Another CRF toolkit
//
//  $Id: feature_index.h 1588 2007-02-12 09:03:39Z taku $;
//
//  Copyright(C) 2005-2007 Taku Kudo <taku@chasen.org>
//
#ifndef CRFPP_FEATURE_INDEX_H_
#define CRFPP_FEATURE_INDEX_H_

//#define _DEBUG

#include <vector>
#include <map>
#include <set>
#include <iostream>
#include <assert.h>
#include "common.h"
#include "scoped_ptr.h"
#include "feature_cache.h"
#include "node.h"
#include "freelist.h"
#include "mmap.h"
#include "darts.h"

#include "../Parsing/Utils.h"
#include "../Dataset.h"

using namespace std;
namespace CRFPP {

class Tagger;
class TaggerImpl;
class SemiTagger;
class SemiJointTagger;

class FeatureIndex {
protected:
  unsigned int       maxid_; //number of features
  double             *alpha_; //crf parameters
  float              *alpha_float_; //redundant

  double              cost_factor_;

  std::vector<const char*>  y_;

  FeatureCache        feature_cache_; //for label and unlabeled
  FeatureValueCache   feature_value_cache_; //for label and unlabeled
  FreeList<char>      char_freelist_;

  // For CRF
  //scoped_array< FreeList<Path> > path_freelist_;
  //scoped_array< FreeList<Node> > node_freelist_;
  scoped_array<NodeCache>  node_freelist_;
  scoped_array<PathCache>  path_freelist_;

  scoped_array< SegmentNodeCache > segnode_freelist_;

public:
   virtual int addID(const char *key, int fsize)=0;
   virtual int getID(const char *) = 0;
   virtual int getfID(const char *key) = 0; //get id without adding
  //const char *get_index(char *&, size_t, const TaggerImpl &);
  whatlog             what_;
  size_t              thread_num_;

public:
  bool      loss_aware;
  int       nony;
  int       non_class;
  int       sparse_feature_index;

  int       polar_ysize_;
  int       intensity_ysize_;

  map<string, int>   lb2i;
  map<int, string>   i2lb;

  vector<string>   polar_ystr;
  vector<string>   intensity_ystr;

  int       word_vec_size;
  int       embedding_label_size;

  vector<vector<float> > word_embeddings;
  map<string, int>    word_lookup_table; //word to embedding index

  map<string, int> dict_features;
  map<string, vector<string> > feature_map;

  map<pair<int,int>,int> transition_states;

  // for debugging
  vector<string>    feature_vec;
  // for the single edge feature?
  int               edgeid_;

  static const unsigned int version = MODEL_VERSION;


  int       windowsize;
  bool      zero_order;
  int       max_span_length;

public:
  inline void set_alpha(int index, double c) { alpha_[index] = c; }
  inline double get_alpha(int index) { return alpha_[index]; }
  virtual bool open(const char*) = 0;

  void LoadEmbeddings(string infilename);
  void InitEmbeddingFeatures();
  void ClearWordEmbeddings();

  void ClearFeatureMap();
  void LoadFeatureMap(string filename);
  // clear everything
  void clear();

  void init() {
     //path_freelist_.reset(new FreeList<Path> [thread_num_]);
     //node_freelist_.reset(new FreeList<Node> [thread_num_]);
     //for (size_t i = 0; i < thread_num_; ++i) {
     //  path_freelist_[i].set_size(8192 * 16);
     //  node_freelist_[i].set_size(8192);
     //}
	  path_freelist_.reset(new PathCache [thread_num_]);
	  node_freelist_.reset(new NodeCache [thread_num_]);
	  /*for (int i = 0; i < thread_num_; ++i) {
		  PathCache p;
		  path_freelist_.push_back(p);
		  NodeCache n;
		  node_freelist_.push_back(n);
	  }*/

	 segnode_freelist_.reset(new SegmentNodeCache [thread_num_]);
  }

  void freelist() {
	  for(int thread_id=0; thread_id<thread_num_; thread_id++) {
		  path_freelist_[thread_id].free();
		  node_freelist_[thread_id].free();

		  segnode_freelist_[thread_id].free();
	  }
  }

  void   clear_alpha() { for(size_t i=0; i<maxid_; i++) alpha_[i] = 0.0; }
  size_t size() const  { return maxid_; }

  int get_lb2i(string str) {
	  if (lb2i.find(str) == lb2i.end()) return -1;
	  return lb2i[str];
  }
  //string fstr (int id) { return fid2str[id2fid[id]]; }

  size_t ysize() const { return y_.size(); }

  size_t labelsize() const { return y_.size() + polar_ystr.size() + intensity_ystr.size(); }

  const char* ystr(size_t i) const {
	  if (i < 0 || i >= ysize()) return "ERR";
	  return y_[i];
  }

  int yID(string l) {
	  if(lb2i.find(l) == lb2i.end())
		  return -1;
	  else
		  return lb2i[l];
  }

  int polarID(string l) {
	  for (int i = 0; i < polar_ystr.size(); ++i) {
		  if (polar_ystr[i] == l) return i;
	  }
	  return -1;
  }

  int intensityID(string l) {
		for (int i = 0; i < intensity_ystr.size(); ++i) {
			if (intensity_ystr[i] == l) return i;
		}
		return -1;
	}

  void   set_alpha(double *alpha) { alpha_ = alpha; }

  const float *alpha_float() { return const_cast<float *>(alpha_float_); }
  const double *weight() const { return const_cast<double *>(alpha_); }
  void set_cost_factor(double cost_factor) { cost_factor_ = cost_factor; }
  double cost_factor() const { return cost_factor_; }

  void SetLabels(set<string> labels);

  void calcCost(Node *);
  void calcCost(Path *);

  const char* getFeature(int cur, int row, int col, TaggerImpl *tagger);

  void buildCRFTagger(TaggerImpl *);
  bool buildCRFFeatures(TaggerImpl *, bool train);

  void LoadDictFeatures(string filename);

  void buildSemiCRFTagger(SemiTagger *tagger);
  bool buildSemiCRFFeatures(SemiTagger *tagger, bool train);

  const char* what() { return what_.str(); }
  char *strdup(const char *);

  explicit FeatureIndex(): maxid_(0), alpha_(0), alpha_float_(0), word_vec_size(0), sparse_feature_index(0),
                           cost_factor_(1.0), nony(0), non_class(0), edgeid_(-1),
                           thread_num_(1), windowsize(1), char_freelist_(8192),
                           zero_order(false),
                           max_span_length(1),
                           polar_ysize_(0), intensity_ysize_(0),
                           loss_aware(false)
                           {}
  virtual ~FeatureIndex() {}
};

class EncoderFeatureIndex: public FeatureIndex {
 private:
  //feature str -> feature id, count
  std::map <std::string, std::pair<int, unsigned int> > dic_;

  int getID(const char *);
  bool openTagSet(const char *);
  int  addID(const char *key, int fsize);

 public:
  explicit EncoderFeatureIndex(size_t n=1) {
	  thread_num_ = n;
      init();
  }

  int getfID(const char*);

  bool open(const char*);

  bool opendata(vector< vector<string> > data);
  bool save(const char *, bool);

  void shrink(size_t) ;

  void BuildFeatureVec() {
	  feature_vec.clear();
	  feature_vec.resize(maxid_);
	  size_t ysize_ = ysize();
	  for (map<string, pair<int, unsigned int> >::iterator it = dic_.begin(); it != dic_.end(); ++it) {
		  string fstr = it->first;
		  int fid = it->second.first;
		  if (fstr[0] == 'B') {
			  for (size_t yi = 0; yi < ysize_; ++yi) {
				  for (size_t yj = 0; yj < ysize_; ++yj) {
					  feature_vec[fid + yi*ysize_+yj] = fstr + "_" + Utils::int2string(yi) + "_" + Utils::int2string(yj);
				  }
			  }
		  } else {
			  for (size_t y = 0; y < ysize_; ++y) {
				  feature_vec[fid+y] = fstr + "_" + Utils::int2string(y);
			  }
		  }
	  }
  }
};

class DecoderFeatureIndex: public FeatureIndex {
private:
	Mmap <char> mmap_;
	Darts::DoubleArray da_;

  int getID(const char *);
  int  addID(const char *key, int fsize) { return -1;}
public:
  int getfID(const char*);
  explicit DecoderFeatureIndex() {
	  init();
  }
  bool open(const char *); //open model
};
}

#endif
