//
//  CRF++ -- Yet Another CRF toolkit
//
//  $Id: tagger.h 1588 2007-02-12 09:03:39Z taku $;
//
//  Copyright(C) 2005-2007 Taku Kudo <taku@chasen.org>
//
#ifndef CRFPP_TAGGER_H__
#define CRFPP_TAGGER_H__

#include <iostream>
#include <vector>
#include "param.h"
#include "scoped_ptr.h"
#include "feature_index.h"
#include "viterbi_decoder.h"

#include "../Sentence.h"

using namespace std;

namespace CRFPP {

static inline double toprob(Node *n, double Z) {
	return std::exp(n->alpha + n->beta - n->cost - Z);
}

static inline double toprob(double cost, double Z) {
	return std::exp(cost - Z);
}

class Tagger {
public:
	Tagger() {};
	virtual ~Tagger(){};
  	int taggerid;
public:
  	size_t                   feature_id_;
	size_t                   feature_value_id_;
	unsigned short           thread_id_;

	FeatureIndex             *feature_index_;

  public:
  	set<int> feature_list;

  	Sentence *sent;

  	std::vector<unsigned short int>  answer_;
  	std::vector<unsigned short int>  result_;

  	virtual  double   gradient(double *)=0;
  	virtual  int      eval()=0;
  	virtual  string   toString()=0;
  	virtual  bool     Inference()=0;
  	virtual  size_t   size() const=0;
  	virtual  bool     next()=0;

   void   set_feature_id(size_t id) { feature_id_  = id; }
   size_t feature_id() const { return feature_id_; }
   void   set_feature_value_id(size_t id) { feature_value_id_  = id; }
   size_t feature_value_id() const { return feature_value_id_; }

   void   set_thread_id(unsigned short id) { thread_id_ = id; }
   unsigned short thread_id() const { return thread_id_; }

};

class TaggerImpl : public Tagger {
  enum { TEST, LEARN };
	  //enum {MINRISK, VITERBI};
  private:
      size_t                            ysize_;

  public:
      std::vector<std::vector <const char *> > x_;
      unsigned int mode_   : 2;
      unsigned int vlevel_ : 3;

      bool          zero_order;

	  double        costC;
      double                   Z_;

      int start_offset;
      int end_offset;

      map<int, int>   label_count;
      vector<vector<string> > connective_features;

      std::vector<std::vector <Node *> > node_;

      whatlog what_;
      string_buffer os_;

      ViterbiDecoder viterbi_decoder;

public:
      explicit TaggerImpl(): mode_(TEST), vlevel_(0),
                             ysize_(0), Z_(0),
                             costC(5), zero_order(false), start_offset(0), end_offset(0) {
    	  	sent = NULL;
    	  	feature_id_ = 0;
    	  	feature_value_id_ = 0;
    	  	thread_id_ = 0;
    	  	feature_index_ = 0;
      }
	  ~TaggerImpl() {
		  /*for (int i = 0; i < node_.size(); ++i) {
			  for (int j = 0; j <node_[i].size(); ++j) {
				  delete node_[i][j];
				  node_[i][j] = NULL;
			  }
		  }*/
	    //close();
	  }

public:
	  double segment_prob(int start, int end);
	  bool next();

	  void calcCost(Node *n);
	  void recomputeCostwithLoss();
	  double SpanLoss(int start, int end, int y, int y_hat);
	  void SegmentInferenceWithLoss();

	  void InferencewithNonDuals(vector<double> &u, double sign);
	  void Decoding();

	  bool SpanToTagger(Sentence *s, Span &en, CRFPP::FeatureIndex *feature_index, bool train);
	  bool SentenceToTagger(Sentence *sent, CRFPP::FeatureIndex *feature_index, bool train);

	  void PrintMarginals();
	  void PrintParameters();

	  bool Inference();
	  void InferencewithDuals(vector<vector<double> > &u, double sign);

	  bool read(std::istream *);
      bool readinst(vector<string> instlines);
      bool add2(size_t, const char **, bool);
      bool add(const char*); //read in a line (token)

      //for inference
      bool testopen(string modelfile);
      bool buildTestFeatureIndex();
      bool parse_stream(std::istream *, std::ostream *);
      bool parse();

      bool outputPrediction(std::ostream *os, vector<string>& results);
      bool resulttoString( vector<string>& results);
      string toString();

      int  eval();
      void  eval(map<pair<int, int>, int> &confusion_matrix);

      //clear data structure
      bool clear();

      bool open(FeatureIndex *);
      void close();

      bool    shrink(bool train);
      void    buildLattice(); //recomputecost + recomputeconstraintcost

      void    recomputeCost();
      void    computerZ();
      double  gradient(double *expected);

      void    recomputeConstraintCost();
      void    forwardbackwardwithConstraints();

      double     labelMarginal();

      void label_evaluate(); //evalute on basic features
      void evaluate();

	  int get_lb2i(string str) {return feature_index_->yID(str);}
	  const char* yname(size_t i) const   { return feature_index_->ystr(i); }
	  size_t y(size_t i)  const    { return result_[i]; }
	  const char* y2str(size_t i) const      { return yname(result_[i]); }
	  const char*  x(size_t i, size_t j) const { return x_[i][j]; } //return feature j of token i
	  vector <const char *>  x(size_t i) { return x_[i];} //return feature list of token i
	  size_t answer(size_t i) const { return answer_[i]; }

      Node  *node(size_t i, size_t j) const { return node_[i][j]; }
      Path  *path(size_t cur, size_t i, size_t j) {return node_[cur][i]->lpath[j];}
      void   set_node(Node *n, size_t i, size_t j) { node_[i][j] = n; }

      size_t       size() const { return x_.size(); } //seq_len
      size_t       ysize() const { return ysize_; }
      //const float *weight_vector() const { return feature_index_->alpha_float(); }
      bool         empty() const { return x_.empty(); }


    double span_cost(int start, int end, size_t y);
    void span_prob(int start, int end, vector<double> &probs);

    double emission_cost(size_t i, size_t j) const { return node_[i][j]->cost; }
	double next_transition_cost(size_t i, size_t j, size_t k);
	double prev_transition_cost(size_t i, size_t j, size_t k);
	double best_cost(size_t i, size_t j) const { return node_[i][j]->bestCost; }

	unsigned int vlevel() const { return vlevel_; }
	float cost_factor() const { return feature_index_->cost_factor(); }

	//use in nbest inference
	double  prob() const { return std::exp(- viterbi_decoder.cost_ - Z_); }

	void set_vlevel(unsigned int vlevel) { vlevel_ = vlevel;}
	void set_cost_factor(float cost_factor) {
	if (cost_factor > 0)
	  feature_index_->set_cost_factor(cost_factor);
	}

	const char* what() { return what_.str(); }

	//Compute statistics for Sampling
	double stateEnergy(int pos, unsigned int k, vector<unsigned short int> y);
	double ComputeEnergyChange(vector<unsigned short int> y, int var_index,
		  unsigned short int old_state, unsigned short int new_state);
	double EvaluateEnergy(vector<unsigned short int> y);
	void getconditional(int pos, vector<unsigned short int> y, vector<double>& p, double temp);
  };
}

#endif
