#pragma once
#include <iostream>
#include <vector>
#include <queue>
#include <fstream>
#include <set>
#include <string>
#include <algorithm>
#include <assert.h>
#include <map>

#include "./Parsing/Utils.h"
#include "./Parsing/PennTree.h"
#include "./Parsing/GraphNode.h"
#include "./Parsing/Dictionary.h"

using namespace std;

class Entity
{
public:
	int sentid;
	int start;
	int end;

	int id; //unique id for the mention

	double value;
	string label;
	int index;

	vector<string> features;
	vector<double> probs;

public:
	Entity() {sentid = -1; id = -1; start = -1; end= -1; label = ""; value = 1; index = 0;}
	Entity(int s, int e, string l="") {sentid = -1;id = -1;start = s; end = e; label = l; value = 1; index = 0;}

	Span GetSpan() { return Span(start, end, label); }

	bool intersect(Entity en)
	{
		if(isEqual(en))
			return false;
		if(overlap(en) > 0)
			return true;
		return false;
	}

	int overlap(Entity en)
	{
		int n = 0;
		for(int i=start; i<=end; i++)
		{
			if(i>=en.start && i<=en.end)
				n += 1;
		}
		return n;
	}

	double f1(Entity en)
	{
		//this is true entity
		double n = overlap(en);
		double p = (double)n/en.len();
		double r = (double)n/len();
		if(p == 0 || r ==0)
			return 0;
		return (double)2*p*r/(p+r);
	}

	bool isEqual(Entity en)
	{
		if(start == en.start && end == en.end)
			return true;
		else
			return false;
	}

	bool isContain(int i)
	{
		if(i>=start && i<=end)
			return true;
		else
			return false;
	}

	string toString();
	string toString(vector<string> words)
	{
		string str = "";
		for(int i=start; i<=end; i++)
			str += words[i] + " ";
		return str.substr(0,str.size()-1);
	}

	int len() {if(start == -1 || end == -1) return 0; else return end-start+1;}
	bool inRange(int s, int e)
	{
		if(s<=start && e>=end)
			return true;
		return false;
	}
};

class Sentence
{
public:
	Sentence() : penntree(0), G(0), dep_tree(0) {};
	~Sentence();
public:
	string sent; //original sentence

	vector<string> words;
	vector<string> lemmas;
	vector<string> postags;
	vector<string> token_labels;

	vector<Span> entities;

	int sent_id;
public:
	PennTree* penntree;
	DependencyGraph* G;
	DependencyTree *dep_tree;
public:
	string ToString();

	string toCRFstring(set<string> &labelmap);

	bool parseCRFSent(string sentstr);

	void genSpanFeatures(int start, int end, vector<string>& features);
	void genSegmentSyntacticFeatures(int start, int end, vector<string>& features);
	void genCRFWindowFeatures(int start, int end, vector<string>& features);
	void genSegmentBoundaryFeatures(int offset, int cur, vector<string> &features);

	bool EmptySpan(int start, int end);
	void genSegmentFeatures(int start, int end, vector<string>& features);
	void genBoundaryFeatures(int window, int cur, vector<string> &features);
	void genBasicFeatures(int cur, vector<string> &features);

	void genParseFeatures(int cur, vector<string> &features);
	void genDepFeatures(int i, vector<string> &features);

	void genEmbeddingFeatures(int start, int end,
			map<string, int> &lookup_table,
			vector<vector<float> > &word_vec,
			vector<double>& feature_values);

	void SetSentenceID(int i) {sent_id = i;}
	int SentenceID() {return sent_id;}

	void buildPennTree(string parse_tree);

	bool buildDepGraph(vector<string> &lines);
	bool buildDepTree(vector<string> &lines);
	bool buildPennTree(vector<string> &lines);
	void buildDiscourseTree(vector<string> &lines);

	void parseSent(string sentstr);

	int size() {return words.size();}

	void printSentence()
	{
		for(int i=0; i<words.size(); i++)
			cout<<words[i]<<" ";
		cout<<endl;
	}

	string GetContext(int start, int end);
	string GetSpanText(int start, int end);

	void GenPathFeatures(int pi, int ci, vector<string> &features);
	void genUnigramFeatures(int start, int end, vector<string> &features);
};

