#include "SemiCRFLearner.h"
#include "CRFLearner.h"
#include "Logger.h"
#include "Config.h"
#include <cstdio>

using namespace std;

void LoadParam(Config props, Param &param) {
	param.use_parse = props.GetBoolProperty("use_parse");
	param.freq = props.GetIntProperty("freq");
	param.threadnum = props.GetIntProperty("threadnum");
	param.max_length = props.GetIntProperty("max_length");
	param.regularizer = props.GetDoubleProperty("regularizer");
	param.lossaware = props.GetBoolProperty("lossaware");

	param.modeltype = props.GetProperty("modeltype");
	param.modelfile = props.GetProperty("modelfile");
	param.resultfile = props.GetProperty("resultfile");
	param.labelscheme = props.GetProperty("labelscheme");

	param.lexicalfeatures = props.GetProperty("lexfeaturefile");
	param.embeddingfile = props.GetProperty("embeddingfile");
}

Dataset *LoadData(string datafile) {
	Dataset *data = new Dataset();
	data->LoadCRFData(datafile);

	return data;
}

void Train(Dataset *traindata, Config props) {
	Param param;
	LoadParam(props, param);

	CRFPP::FeatureIndex *feature_index = NULL;

	set<string> ylabels;
	traindata->CollectTaggingLabels(ylabels);
	cout<<"Training labels:"<<endl;
	for (set<string>::iterator it = ylabels.begin(); it != ylabels.end(); ++it) {
		cout<<*it<<" ";
	}
	cout<<endl;

	traindata->CollectEntities(param.labelscheme);

	// Building feature index
	feature_index = new CRFPP::EncoderFeatureIndex(param.threadnum);
	feature_index->SetLabels(ylabels);


	feature_index->zero_order = param.zeroorder;
	feature_index->windowsize = param.windowsize;

	if (param.lexicalfeatures != "")
		feature_index->LoadFeatureMap(param.lexicalfeatures);

	if (param.embeddingfile != "") {
		// read embeddings
		feature_index->LoadEmbeddings(param.embeddingfile);
		feature_index->embedding_label_size = feature_index->ysize();
		feature_index->InitEmbeddingFeatures();
	}

	if (param.modeltype == "crf") {
		CRFLearner crf_extractor;
		crf_extractor.initParam(param);
		crf_extractor.trainset = traindata;
		crf_extractor.InitializeEncoder(feature_index);

		crf_extractor.BuildSentenceTagger();
		crf_extractor.TrainEncoder(param.modelfile); //sgd, highorder
	} else if (param.modeltype == "semicrf") {
		feature_index->loss_aware = param.lossaware;
		feature_index->max_span_length = param.max_length;

		SemiCRFLearner semicrf_extractor;
		semicrf_extractor.initParam(param);
		semicrf_extractor.trainset = traindata;
		semicrf_extractor.InitializeEncoder(feature_index);

		semicrf_extractor.BuildTagger();
		semicrf_extractor.TrainEncoder(param.modelfile); //sgd, highorder
	}

	delete feature_index;
	feature_index = NULL;
}

void Inference(Dataset *testdata, Config props) {
	Param param;
	LoadParam(props, param);

	CRFPP::FeatureIndex *feature_index = NULL;
	feature_index = new CRFPP::DecoderFeatureIndex();
	feature_index->open(param.modelfile.c_str());

	testdata->CollectEntities(param.labelscheme);

	if (feature_index != NULL) {
		if (param.lexicalfeatures != "")
			feature_index->LoadFeatureMap(param.lexicalfeatures);

		if (param.embeddingfile != "") {
			// read embeddings
			feature_index->LoadEmbeddings(param.embeddingfile);
			feature_index->embedding_label_size = feature_index->ysize();
			feature_index->InitEmbeddingFeatures();
		}
	}

	if (param.modeltype == "crf") {
		CRFPredictor chain_predictor;
		chain_predictor.InitializeDecoder(feature_index);
		chain_predictor.nbest = param.nbest;

		chain_predictor.testset = testdata;
		chain_predictor.BuildSentenceTagger();

		chain_predictor.eval_param.overlap = props.GetIntProperty("eval_overlap");
		chain_predictor.eval_param.key = props.GetProperty("eval_key");
		chain_predictor.eval_param.scheme = props.GetProperty("eval_scheme");

		chain_predictor.TaggingPrediction(param.resultfile);
	} else if (param.modeltype == "semicrf") {
		SemiCRFPredictor semicrf_predictor;
		semicrf_predictor.InitializeDecoder(feature_index);
		semicrf_predictor.nbest = param.nbest;

		feature_index->max_span_length = param.max_length;

		semicrf_predictor.testset = testdata;
		semicrf_predictor.BuildTagger();

		semicrf_predictor.eval_param.overlap = props.GetIntProperty("eval_overlap");
		semicrf_predictor.eval_param.key = props.GetProperty("eval_key");
		semicrf_predictor.eval_param.scheme = props.GetProperty("eval_scheme");

		semicrf_predictor.TaggingPrediction(param.resultfile);
	}

	delete feature_index;
	feature_index = NULL;
}

int main(int argc, char **argv)
{
	if (argc <= 1) return 0;

	string configfile = argv[1];
	Config props(configfile);

	string runtype = props.GetProperty("runtype");

	if (runtype == "train-only" || runtype == "both") {
		string trainfile = props.GetProperty("trainfile");
		if (trainfile != "") {
			Dataset *traindata = LoadData(trainfile);
			Train(traindata, props);

			delete traindata;
			traindata = NULL;
		}
	}

	if (runtype == "test-only" || runtype == "both") {
		string testfile = props.GetProperty("testfile");
		if (testfile != "") {
			Dataset *testdata = LoadData(testfile);
			Inference(testdata, props);

			delete testdata;
			testdata = NULL;
		}
	}

	return 0;
}
